//
//  Ansi-C written for Microchip's PIC-X compiler.

#include <xc.h>
#include <pic12f683.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>


// INSTRUCTIONS TO COMPILER     functions used to manipulate and test bits
#define _XTAL_FREQ                  8000000
#define  test_bit(var, bit)   ((var) & (1 <<(bit)))      
#define  set_bit(var, bit)    ((var) |= (1 << (bit)))
#define  clear_bit(var, bit)    ((var) &= ~(1 << (bit)))


// CONFIGURATION BYTES
#pragma config FOSC = INTOSCIO  // Oscillator Selection bits (INTOSCIO oscillator: I/O function on RA4/OSC2/CLKOUT pin, I/O function on RA5/OSC1/CLKIN)
#pragma config WDTE = OFF       // Watchdog Timer Enable bit (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config MCLRE = OFF      // MCLR Pin Function Select bit (MCLR pin function is digital input, MCLR internally tied to VDD)
#pragma config CP = OFF         // Code Protection bit (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Code Protection bit (Data memory code protection is disabled)
#pragma config BOREN = OFF      // Brown Out Detect (BOR disabled)
#pragma config IESO = OFF       // Internal External Switchover bit (Internal External Switchover mode is disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enabled bit (Fail-Safe Clock Monitor is disabled)

// CONSTANTS
#define false						0
#define true						1
#define pulse_dur_2400bps           378   //in microseconds. Tuned to produce the right pulse width. Must produce  a 417us pulse
#define pulse_dur_9600bps           90    //in microseconds. Tuned to produce the right pulse width. Must produce  a 104us pulse
#define ctlr_serial_rx_pin          0     // Output pins definition
#define coil_pulse_pin              1     //     "
#define serial_rx_pin               2     //     "
#define lcd_serial_tx_pin           5     //     "

// GLOBAL VARIABLES
unsigned char i; //Used as counter
unsigned char j; //Used as counter
unsigned char rx_buff[25];    // The received serial data buffer
unsigned char rx_buff_pos;    // A position flag for reading/writing the data buffer
unsigned char rx_char;        // A received character
bool char_removed;            // Boolean flag

//******************************************************************************************************

// Interrupt Service Routine

//void interrupt( void )
//{
//}

// End of interrupt routine

//================================================================

// Software Tx-Only USART, sends one character.
//The code for the TX part of the USART is simplified in that it generates 10 bits of serial data with no parity bit.

void putch(unsigned char tx_char) // must be called "putch" in order to work with the printf command
// This is a software tx uart
{
    unsigned short mask = 1;
    unsigned char i;
    unsigned short txdata;

    txdata = tx_char << 1; // shift left by one the ASCII character to send. Left with 7 bits.
    // start bit (0) + 7 data bits + parity bit (0)
    for (i = 0; i < 9; i++) // lsb first
    {
        if (mask & txdata) set_bit(GPIO, lcd_serial_tx_pin);  // Send one bit
        else clear_bit(GPIO, lcd_serial_tx_pin);              //    "
        mask <<= 1; // Shift mask left by 1
        __delay_us(pulse_dur_9600bps); // wait one bit width duration
    }
    set_bit(GPIO, lcd_serial_tx_pin); // add stop bit (1)
    __delay_us(pulse_dur_9600bps); // wait one bit width duration for stop bit
    __delay_ms(1); // Pause between characters
}

// Software Rx-Only USART, receives one character.
char soft_rx_uart(int pin)
{
    unsigned char c, bitno;
    c = 0;
    while (test_bit(GPIO, pin)); // wait for start bit
    __delay_us(pulse_dur_2400bps / 2); // to be sure its not a glitch sample half-way into bit
    if (test_bit(GPIO, pin)); // was just noise
    else for (bitno = 0; bitno < 8; bitno++) 
    {
            __delay_us(pulse_dur_2400bps);
            c = (c >> 1) | (test_bit(GPIO, pin) << (7 - pin)); // test_bit returns 0 or 00000100 if GP2 is set (does not return 1). Not a boolean function.
            if (test_bit(GPIO, pin)) set_bit(GPIO, 4);
            else clear_bit(GPIO, 4);
    }
    __delay_us(pulse_dur_2400bps); // let the stop bit pass
    __delay_us(pulse_dur_2400bps / 2); // complete the wait for the second half of the stop bit
    return c;
}

// A custom conversion from strign to decimal
unsigned char cvrt_str_to_dec(unsigned char char1, unsigned char char2)
{
    return (char1 - 48) * 10 + (char2 - 48);
}

// A routine to clear the rx buffer
void clr_rx_buff(void)
{
        for (i = 0; i < 25; i++) rx_buff[i] = 0;
        rx_buff_pos = 0;
}

void main()
{
    // Initialization stage
    INTCON      = 0b00000000; // disable interrupts
    IOC         = 0b00000000; // disable GPIO interrupts on change
    PIE1        = 0b00000000; // disable various interrupts
    OPTION_REG  = 0b00001111; // weak pullups global enable,
    WPU         = 0b00110111; // individual weak pullups enabled
    OSCCON      = 0b01110001; // select internal oscillator and set it to 8Mhz
    T1CON       = 0b10110111; // disable timer1 and set prescaler div. by 8
    T2CON       = 0b01111010; // disable timer2 and set prescaler div. by 16 and postscaler div. by 16
    CCP1CON     = 0b00000101; // Set Capture mode, every rising edge
    ADCON0      = 0b00000000; // disable A/D converter
    ANSEL       = 0b00000000; // Disable analog assignment to pins
    CMCON0      = 0b00010111; // Disable comparator, use internal reference, invert output.
    CMCON1      = 0b00000000; // comparator gating timer 1 asynchronously
    VRCON       = 0b00100001; // Disable internal comparator Vref, use low range, set Vref to 0.15V
    WDTCON      = 0b00010100; // Assign the right prescaler value to the watchdog timer. Keep WTD off for now.
    GPIO        = 0b00100000; // Set output pins initially before enabling them.
    TRISIO      = 0b00001111; // GPIO first disabled, but assigned as follows:
                            //	5-output, LCD serial data
                            //	4-input, available
                            //	3-input only, no weak pullup. available
                            //	2-input, Serial Data Rx
                            //	1-input, coil pulse to advance clock on rising edge
                            //	0-input, serial data from Pendulum Control PIC
    //Main loop
    int hours = 0;      // Initialize variables
    int minutes = 0;
    int seconds = 0;
    // Initialize LCD
    __delay_ms(250);     // Wait for one second
    __delay_ms(250);
    __delay_ms(250);
    __delay_ms(250);
    printf("CS%c%c", 0, 10); // turn off cursor
    printf("STCR%c%c%c%c%c%c%c",20,4,128,192,148,212,10);   // Set LCD as a 20 x 4 display
    clr_rx_buff();      // erase serial receive buffer
    rx_buff_pos = 0;
    char_removed = false;

    while (true)   // Main loop that increments the clock and receives the data
    {
        for (hours = 0; hours < 24; hours = hours + 1)
        {
            for (minutes = 0; minutes < 60; minutes = minutes + 1)
            {
                for (seconds = 0; seconds < 60; seconds = seconds + 1)
                {
                    while (!test_bit(GPIO, coil_pulse_pin)) // wait for coil pulse to come in. In the meantime receive the serial ports data
                    {
                            // Receive the data sent by the user
                        if (!test_bit(GPIO, serial_rx_pin)) // if a start bit arrives on the user serial port
                        {
                            rx_char = soft_rx_uart(serial_rx_pin); // read that character
                            if (rx_char != 0) // A 0 would mean a glitch on the Rx port
                            {
                                rx_buff[rx_buff_pos] = rx_char;   // Transfer received character to Rx buffer
                                rx_buff_pos = rx_buff_pos + 1;    // increment buffer position
                                if (rx_buff_pos == 8)             // if we are in 8th position
                                {
                                    if ((rx_buff[0] == 'S')) // Set "S" command received   && (rx_buff[7] == 0xA)  // Is first character received S and last character LF?
                                    {     // yes, this is the time change command
                                        hours = cvrt_str_to_dec(rx_buff[1], rx_buff[2]);     // Update time variables using the recived data
                                        minutes = cvrt_str_to_dec(rx_buff[3], rx_buff[4]);
                                        seconds = cvrt_str_to_dec(rx_buff[5], rx_buff[6]);
                                    }
                                    clr_rx_buff();   // erase Rx buffer
                                }
                            }
                        }
                            // Receive the data sent by the pendulum controller and display it.
                        if (!test_bit(GPIO, ctlr_serial_rx_pin)) // if a start bit arrives on the controller serial port
                        {
                            rx_char = soft_rx_uart(ctlr_serial_rx_pin); // read that character
                            if ((rx_char != 0))  // A 0 would mean a glitch on the Rx port
                            {                                           
                                rx_buff[rx_buff_pos] = rx_char;  // Save character into Rx buffer
                                if ((rx_buff_pos == 16) && (char_removed == false))    // Time to remove one of the characters to limit the test to 20 characters?
                                {
                                    rx_buff_pos = rx_buff_pos - 1;  // this removes character 16 because it is always 0 and there is not enough characters available on LCD.
                                    char_removed = true;    // Set the flag to "character removed"
                                }
                                rx_buff_pos = rx_buff_pos + 1; // increase Rx buffer character position

                                if ((rx_char == 10) || (rx_buff_pos == 21))  // Has end of received string been reached (LF received) ?
                                {
                                    rx_buff_pos = 0;  // re-initialize Rx buffer for next string
                                    printf("TP%c%c%c", 0, 1, 10); // position cursor to (1,0)
                                    printf("TT%s%c",rx_buff ,10); // rx_buff  send received raw data, including \n and delimited by null terminator
                                    clr_rx_buff();  // Clear the Rx buffer
                                    char_removed = false;   // clear the removed character flag
                                }
                            }
                        }
                    }
                    // Here a coil pulse has been received, must update LCD display with time
                    printf("TP%c%c%c", 0, 0, 10); // position cursor to (0,0)
                    printf("TT    %2.2d:%2.2d:%2.2d UTC%c", hours, minutes, seconds, 10); // send time string
                    while (test_bit(GPIO, coil_pulse_pin)); // wait for coil pulse to fall to debounce
                }
            }
        }
    }
}

